from os import path
import math
from copy import deepcopy as copy
import bpy
import bmesh
from mathutils import Matrix


class Mesh(object):
    
    sb = ""
    name = ""
    
    def serialize(self,
              obj: bpy.types.Object,
              obj_mat: Matrix,
              depsgraph: bpy.types.Depsgraph,
              EX_GLOBAL_MATRIX: Matrix,
              EX_ARMATURE=False,
              EX_ANIMATION=False,
              EX_APPLY_MODIFIERS=True):
        
        # get the mesh of the object
        obj_eval = obj.evaluated_get(depsgraph)
        mesh = obj_eval.to_mesh(preserve_all_data_layers=True, depsgraph=depsgraph)
        
        # set the name
        self.name = path.normpath(mesh.name)
        
        # scale it according to global scale matrix
        # mesh.transform(EX_GLOBAL_MATRIX @ obj_mat)
        mesh.transform(EX_GLOBAL_MATRIX)
        # mesh.transform(obj_mat)
        
        # If negative scaling, we have to invert the normals...
        if obj_mat.determinant() < 0.0:
            mesh.flip_normals()
        
        if len(mesh.uv_layers) == 0 and obj.type in {"MESH"}: # generating UV map if none found
            for ob in bpy.data.objects:
                ob.select_set(False)
            bpy.context.view_layer.objects.active = obj
            obj.select_set(True)
            bpy.ops.object.mode_set(mode="EDIT", toggle=False)
            if len(mesh.polygons) > 10000: # let's say anything above 10k is big.
                bpy.ops.uv.reset()
            else:
                bpy.ops.uv.unwrap() # this would be way too slow for complicated or very big geometry
            bpy.ops.object.mode_set(mode="OBJECT", toggle=False)
            obj_eval = obj.evaluated_get(depsgraph)
            mesh = obj_eval.to_mesh(preserve_all_data_layers=True, depsgraph=depsgraph)
        
        # bmesh conversion for triangulation
        bm = bmesh.new()
        bm.from_mesh(mesh)
        # bmesh.ops.scale(bm, vec=EX_GLOBAL_MATRIX.decompose()[2], verts=bm.verts)
        bmesh.ops.triangulate(bm, faces=bm.faces)
        bm.to_mesh(mesh)
        bm.free()
        
        # calculate tangents for the mesh (ogre needs these for some odd reason)
        mesh.calc_tangents()
        
        # init string list and indent string
        sl = []
        ind = "\t"
        
        # Roots of the XML graph
        sl.append("<mesh>\n")
        sl.append(f"{ind}<sharedgeometry vertexcount=\"__VERTEXCOUNT__\">\n") # This'll have to be replaced later
        sl.append(f"{ind}{ind}<vertexbuffer binormals=\"{True}\" colours_diffuse=\"{False}\" normals=\"{True}\" positions=\"{True}\" tangent_dimensions=\"{3}\" tangents=\"{True}\" texture_coords=\"{len(mesh.uv_layers)}\">\n") # lookup the number of UV maps
        
        # yes, index hell incoming
        d_processed_verts = {} # {"idx_vert": [  {"idx_ogre_vert": [[x, y, z], [[u, v], [...]] ]}, {...}  ]}
        d_ogre_faces = {} # {"idx_ogre_face": [idx_mat, idx_ogre_vert1, idx_ogre_vert2, idx_ogre_vert3]}
        d_materials = {} # {"idx_mat": "material.name"}
        
        # collecting material index - name relations
        for idx_mat in range(0, len(mesh.materials)):
            if mesh.materials[idx_mat] is not None:
                d_materials[idx_mat] = mesh.materials[idx_mat].name
            else:
                d_materials[idx_mat] = "none"
        if len(mesh.materials) == 0:
            d_materials[0] = "none"
        
        def near_eq(cur_norl, cur_l_uv, ref_norl, ref_l_uv) -> bool:
            # print(f"INFO - near_eq: {cur_nord} ? {ref_nord}")
            if not math.isclose(cur_norl[0], ref_norl[0], rel_tol=1e-08): return False
            if not math.isclose(cur_norl[1], ref_norl[1], rel_tol=1e-08): return False
            if not math.isclose(cur_norl[2], ref_norl[2], rel_tol=1e-08): return False
            for cuvd, ruvd in zip(cur_l_uv, ref_l_uv):
                if not ((cuvd[0] == ruvd[0]) and (cuvd[1] == ruvd[1])):
                    return False
            return True
        
        # Ready, set...
        io_next_vert = 0 # set vertex count to 0, increase with every vertex to be written
        io_current_face = 0 # same with faces
        
        # Lets GOOO
        for poly in mesh.polygons: # these be tris
            d_ogre_faces[io_current_face] = [poly.material_index]
            for idx_loop in poly.loop_indices: # these be loops
                loop = mesh.loops[idx_loop]
                idx_vert = loop.vertex_index # indexes are important
                
                # mapping
                posl = []
                norl = []
                tanl = []
                binl = []
                l_uv = []
                
                # über simple mapping for speedz
                posl = mesh.vertices[idx_vert].co
                # poly.normal # loop.normal # mesh.vertices[idx_vert].normal ?
                norl = loop.normal
                
                for uvlindex in range(0, len(mesh.uv_layers)):
                    uvlayer = mesh.uv_layers[uvlindex]
                    uvl = [uvlayer.data[idx_loop].uv[0], -uvlayer.data[idx_loop].uv[1]]
                    l_uv.append(uvl)
                
                tanl = loop.tangent
                binl = loop.bitangent
                
                # super important!
                # check if we have already processed a given vertex and reference it if possible.
                found = False
                if idx_vert in d_processed_verts:
                    for idx_ogre_vert in d_processed_verts[idx_vert]:
                        if near_eq(norl, l_uv, d_processed_verts[idx_vert][idx_ogre_vert][0], d_processed_verts[idx_vert][idx_ogre_vert][1]):
                            d_ogre_faces[io_current_face].append(idx_ogre_vert)
                            found = True
                            break
                    if not found:
                        d_processed_verts[idx_vert][io_next_vert] = [norl, l_uv]
                        d_ogre_faces[io_current_face].append(io_next_vert)
                        io_next_vert += 1
                else:
                    d_processed_verts[idx_vert] = {io_next_vert: [norl, l_uv]}
                    d_ogre_faces[io_current_face].append(io_next_vert)
                    io_next_vert += 1
                
                if found:
                    continue
                
                # adding vertex data to the master buffer
                sl.append(f"{ind}{ind}{ind}<vertex>\n")
                sl.append(f"{ind}{ind}{ind}{ind}<position x=\"{posl[0]}\" y=\"{posl[1]}\" z=\"{posl[2]}\" />\n")
                sl.append(f"{ind}{ind}{ind}{ind}<normal x=\"{norl[0]}\" y=\"{norl[1]}\" z=\"{norl[2]}\" />\n")
                for uvl in l_uv:
                    sl.append(f"{ind}{ind}{ind}{ind}<texcoord u=\"{uvl[0]}\" v=\"{uvl[1]}\" />\n")
                sl.append(f"{ind}{ind}{ind}{ind}<tangent x=\"{tanl[0]}\" y=\"{tanl[1]}\" z=\"{tanl[2]}\" />\n")
                sl.append(f"{ind}{ind}{ind}{ind}<binormal x=\"{binl[0]}\" y=\"{binl[1]}\" z=\"{binl[2]}\" />\n")
                sl.append(f"{ind}{ind}{ind}</vertex>\n")
                
            io_current_face += 1
        
        # end of blender internal mesh processing
        # here comes the materials
        # material-assignment by submeshes (submeshes by faces)
        
        # this is just no. but also yes.
        sl[1] = sl[1].replace("__VERTEXCOUNT__", f"{io_next_vert}", 1)
        sl.append(f"{ind}{ind}</vertexbuffer>\n")
        sl.append(f"{ind}</sharedgeometry>\n")
        
        
        # write submeshes
        sl.append(f"{ind}<submeshes>\n")
        for idx_mat in d_materials:
            
            # Collect faces to write
            b32b = "True"
            idxs_of = []
            for idx_ogre_face in d_ogre_faces:
                if d_ogre_faces[idx_ogre_face][0] == idx_mat:
                    idxs_of.append(idx_ogre_face)
            # if len(idxs_of) < 21844:
            #     b32b = "False"
            
            name = d_materials[idx_mat]
            sl.append(f"{ind}{ind}<submesh material=\"{name}\" operationtype=\"triangle_list\" use32bitindexes=\"{b32b}\" usesharedvertices=\"True\">\n")
            
            # write faces data
            sl.append(f"{ind}{ind}{ind}<faces count=\"{len(idxs_of)}\">\n")
            
            for idx_ogre_face in idxs_of:
                sl.append(f"{ind}{ind}{ind}{ind}<face v1=\"{d_ogre_faces[idx_ogre_face][1]}\" v2=\"{d_ogre_faces[idx_ogre_face][2]}\" v3=\"{d_ogre_faces[idx_ogre_face][3]}\" />\n")
            
            sl.append(f"{ind}{ind}{ind}</faces>\n")
            sl.append(f"{ind}{ind}</submesh>\n")
        sl.append(f"{ind}</submeshes>\n")
        
        
        # Assign submeshes to materials
        sl.append(f"{ind}<submeshnames>\n")
        for idx_mat in d_materials:
            name = d_materials[idx_mat]
            sl.append(f"{ind}{ind}<submesh index=\"{idx_mat}\" name=\"{name}\" />\n")
        sl.append(f"{ind}</submeshnames>\n")
        sl.append("</mesh>\n")
        # Aaand we are done.
        
        # Remove temporary mesh.
        # obj_eval.to_mesh_clear()
        
        # pass the completed buffer back to the parent object
        self.sb = "".join(sl)
        # return number of vertices
        return io_next_vert
